# SPDX-FileCopyrightText: 2023-present Oori Data <info@oori.dev>
#
# SPDX-License-Identifier: Apache-2.0

# XXX: I think this test has an __init__.py as part of hatch's
# developer/live environments scheme? If this is the case,
# it's not long for this world. We just use pytest directly.
